#!/usr/bin/python3
# -*- coding:utf-8 -*-

#  Copyright © 2010, 2015-2016  B. Clausius <barcc@gmx.de>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License,  or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import IntEnum
import random

from .common import ffi, print_cdata, printd

ffi.cdef("""
//from bridges.c
#define G_ISLAND        0x0001
#define G_LINEV         0x0002     /* contains a vert. line */
#define G_LINEH         0x0004     /* contains a horiz. line (mutex with LINEV) */
//#define G_LINE          (G_LINEV|G_LINEH)
#define G_MARKV         0x0008
#define G_MARKH         0x0010
//#define G_MARK          (G_MARKV|G_MARKH)
#define G_NOLINEV       0x0020
#define G_NOLINEH       0x0040
//#define G_NOLINE        (G_NOLINEV|G_NOLINEH)

typedef unsigned int grid_type; /* change me later if we invent > 16 bits of flags. */

struct game_ui {
    int dragx_src, dragy_src;   /* source; -1 means no drag */
    int dragx_dst, dragy_dst;   /* src's closest orth island. */
    grid_type todraw;
    int dragging, drag_is_noline, nlines;

    int cur_x, cur_y, cur_visible;      /* cursor position */
    int show_hints;
};

struct game_params {
    int w, h, maxb;
    int islands, expansion;     /* %age of island squares, %age chance of expansion */
    int allowloops, difficulty;
};

struct surrounds { /* cloned from lightup.c */
    struct { int x, y, dx, dy, off; } points[4];
    int npoints, nislands;
};

struct island {
  game_state *state;
  int x, y, count;
  struct surrounds adj;
};

struct game_state {
    int w, h, completed, solved, allowloops, maxb;
    grid_type *grid;
    struct island *islands;
    int n_islands, n_islands_alloc;
    game_params params; /* used by the aux solver. */
//#define N_WH_ARRAYS 5
    char *wha, *possv, *possh, *lines, *maxv, *maxh;
    struct island **gridi;
    struct solver_state *solver; /* refcounted */
};
""")


class UiState:
    G_ISLAND  = 0x0001
    G_LINEV   = 0x0002     #/* contains a vert. line */
    G_LINEH   = 0x0004     #/* contains a horiz. line (mutex with LINEV) */
    G_LINE    = (G_LINEV|G_LINEH)
    G_MARKV   = 0x0008
    G_MARKH   = 0x0010
    G_MARK    = (G_MARKV|G_MARKH)
    G_NOLINEV = 0x0020
    G_NOLINEH = 0x0040
    G_NOLINE  = (G_NOLINEV|G_NOLINEH)
    
    def __init__(self, ui, state):
        self.ui = ui
        self.state = state
        
    grid = lambda self,x,y: self.state.grid[x+y*self.state.w]
    index = lambda self,attr,x,y: getattr(self.state, attr)[x+y*self.state.w]
    lines = lambda self,x,y: self.state.lines[x+y*self.state.w][0]
    gridi = lambda self,x,y: self.state.gridi[x+y*self.state.w]
    isisland = lambda self,x,y: self.grid(x,y) & self.G_ISLAND
    islineh = lambda self,x,y: self.grid(x,y) & self.G_LINEH
    islinev = lambda self,x,y: self.grid(x,y) & self.G_LINEV
    isline = lambda self,adj: (self.islineh(adj.x,adj.y) if adj.dx else self.islinev(adj.x,adj.y))
    ismark = lambda self,x,y: self.grid(x,y) & self.G_MARK
    
    @property
    def cur_xy(self):
        return self.ui.cur_x, self.ui.cur_y
    @cur_xy.setter
    def cur_xy(self, xy):
        self.ui.cur_x, self.ui.cur_y = xy
        

def ensure_valid(uistate):
    def first(w, h):
        for y in range(h):
            for x in range(w):
                if uistate.isisland(x,y):
                    uistate.cur_xy = x, y
                    return
        assert False
    movestr = None
    if not uistate.ui.cur_visible:
        uistate.ui.cur_visible = 1
        movestr = ''
        global values
        values = algorithm.new_values()
    w, h = uistate.state.w, uistate.state.h
    cx, cy = uistate.cur_xy
    if w > cx >= 0 <= cy < h and uistate.isisland(cx, cy):
        return movestr
    first(w, h)
    return ''
    

class AlgorithmBase:
    class Values:  pass
    turns = [(1,0), (0,1), (-1,0), (0,-1)]
    
    @classmethod
    def new_values(self):
        values = self.Values()
        values.direction = 0
        return values
        
    @staticmethod
    def adj_iter(island):
        assert island
        for i in range(island.adj.npoints):
            yield island.adj.points[i]
            
    @classmethod
    def isfreeview(self, uistate, island, adj):
        if adj.dx == 0:
            assert adj.dy != 0
            if adj.dy < 0:
                miny, maxy = island.y - adj.off, island.y
            else:
                miny, maxy = island.y, island.y + adj.off
            for y in range(miny+1, maxy):
                if uistate.islineh(island.x, y):
                    return False
        else:
            assert adj.dy == 0
            if adj.dx < 0:
                minx, maxx = island.x - adj.off, island.x
            else:
                minx, maxx = island.x, island.x + adj.off
            for x in range(minx+1, maxx):
                if uistate.islinev(x, island.y):
                    return False
        return True
        
    @classmethod
    def can_go_forward(self, uistate, island, adj, nx, ny):
        if uistate.isline(adj):
            return True
        # 1. Only walk over bridges
        return False
        # 2. Walk where may be a bridge
        if uistate.ismark(island.x, island.y):
            return False
        if uistate.ismark(nx, ny):
            return False
        return self.isfreeview(uistate, island, adj)
        
    @classmethod
    def check(self, uistate):
        def get_free_bridges(island):
            assert island
            cnt = 0
            for adj in self.adj_iter(island):
                if uistate.isline(adj):
                    cnt += uistate.lines(adj.x, adj.y)
            return max(0, island.count-cnt)
        def adj_info_iter(island):
            assert island
            for adj in self.adj_iter(island):
                if not adj.off:
                    assert not uistate.isline(adj)
                    continue
                adj_lines = uistate.lines(adj.x, adj.y) if uistate.isline(adj) else 0
                if adj_lines or self.isfreeview(uistate, island, adj):
                    adj_island = uistate.gridi(island.x+adj.dx*adj.off, island.y+adj.dy*adj.off)
                    yield adj, adj_island, adj_lines
                    
        x, y = uistate.cur_xy
        island = uistate.gridi(x, y)
        adj_info = list(adj_info_iter(island))
        
        # mark full islands
        adj_bridges = sum(al for a,ai,al in adj_info)
        if adj_bridges == island.count and not uistate.ismark(x, y):
            return 'M{},{}'.format(x, y)
            
        maxb = uistate.state.params.maxb
        # create new bridge
        for adj, adj_island, adj_lines in adj_info:
            if (adj.dx, adj.dy) == self.turns[values.direction]:
            #if True:
                adj_max_bridges = sum(min(al+get_free_bridges(ai), maxb)
                                for a,ai,al in adj_info
                                if (a.dx, a.dy) != (adj.dx, adj.dy))
                if adj_max_bridges + adj_lines < island.count:
                    return 'L{},{},{},{},{}'.format(x, y, adj_island.x, adj_island.y, adj_lines+1)
                elif not adj_lines:
                    adj_info_r = list(adj_info_iter(adj_island))
                    adj_max_bridges = sum(min(al+get_free_bridges(ai), maxb)
                                            for a,ai,al in adj_info_r
                                            if (-a.dx, -a.dy) != (adj.dx, adj.dy))
                    if adj_max_bridges < adj_island.count:
                        return 'L{},{},{},{},{}'.format(x, y, adj_island.x, adj_island.y, 1)
                break#
        else:
            return None
            
        #TODO: check for bridges that would create not connected parts
        #TODO: in allow loop mode check for bridges that would create loops
        return None
        
    @classmethod
    def step(self, uistate, i=0):
        movestr = self.check(uistate)
        if movestr is not None:
            return movestr
        return self.algorithm(uistate, i)
        
        
class RighthandAlgorithm (AlgorithmBase):
    @classmethod
    def algorithm(self, uistate, i):
        if i >= 4:
            return None
        x, y = uistate.cur_xy
        island = uistate.gridi(x, y)
        for adj in self.adj_iter(island):
            nx, ny = x+adj.dx*adj.off, y+adj.dy*adj.off
            if (adj.dx, adj.dy) == self.turns[values.direction] and adj.off and self.can_go_forward(uistate, island, adj, nx, ny):
                uistate.cur_xy = nx, ny
                values.direction = (values.direction+1)%4
                return ''
        values.direction = (values.direction-1)%4
        return self.step(uistate, i+1)
        
        
class TarryAlgorithm (AlgorithmBase):
    class tarry(IntEnum): STOP, LAST, NOGO = 1, 2, 3
    
    @classmethod
    def new_values(self):
        values = super().new_values()
        values.tarry_state = {}
        values.dead_islands = {}
        return values
        
    @staticmethod
    def get_adj_island_pos(x, y, adj):
        return x+adj.dx*adj.off, y+adj.dy*adj.off
        
    @classmethod
    def _isdeadend(self, uistate, x, y, adj):
        nx, ny = self.get_adj_island_pos(x, y, adj)
        if not uistate.ismark(nx, ny):
            return False
        di = values.dead_islands.get((nx, ny), False)
        if di:
            return True
        values.dead_islands[(nx, ny)] = True
        nisland = uistate.gridi(nx, ny)
        for nadj in self.adj_iter(nisland):
            if nadj.dx == -adj.dx and nadj.dy == -adj.dy:
                continue
            if not uistate.isline(nadj):
                continue
            if not self._isdeadend(uistate, nx, ny, nadj):
                values.dead_islands[(nx, ny)] = False
                return False
        return True
        
    @classmethod
    def adj_from_direction(self, island):
        for adj in self.adj_iter(island):
            if (adj.dx, adj.dy) == self.turns[values.direction]:
                return adj
                
    @classmethod
    def isdeadend(self, uistate):
        x, y = uistate.cur_xy
        island = uistate.gridi(x, y)
        adj = self.adj_from_direction(island)
        if adj is None:
            return True
        return self._isdeadend(uistate, x, y, adj)
        
    @classmethod
    def algorithm(self, uistate, i):
        global values
        x, y = uistate.cur_xy
        island = uistate.gridi(x, y)
        direction = values.direction
        cur_tarry_state = values.tarry_state.get((x,y))
        step_done = False
        if cur_tarry_state is None:
            values.tarry_state[(x,y)] = cur_tarry_state = [0,0,0,0]
        # go in direction if possible
        for adj in self.adj_iter(island):
            if (adj.dx, adj.dy) != self.turns[direction]:
                continue
            nx, ny = x+adj.dx*adj.off, y+adj.dy*adj.off
            if adj.off and self.can_go_forward(uistate, island, adj, nx, ny):
                assert cur_tarry_state[direction] != self.tarry.NOGO
                if cur_tarry_state[direction] == self.tarry.STOP:
                    break
                if cur_tarry_state[direction] == self.tarry.LAST:
                    if 0 in cur_tarry_state:
                        break
                cur_tarry_state[direction] = self.tarry.STOP
                if self.isdeadend(uistate):
                    break
                uistate.cur_xy = nx, ny
                step_done = True
                cur_tarry_state = values.tarry_state.get((nx,ny))
                if cur_tarry_state is None:
                    values.tarry_state[(nx,ny)] = cur_tarry_state = [0,0,0,0]
                    cur_tarry_state[(direction+2)%4] = self.tarry.LAST
            else:
                cur_tarry_state[direction] = self.tarry.NOGO
            break
        else:
            cur_tarry_state[direction] = self.tarry.NOGO
        # select a new direction if possible
        for terry_value in 0, self.tarry.LAST:
            if terry_value in cur_tarry_state:
                #values.direction = cur_tarry_state.index(terry_value)
                values.direction = random.choice([i for i,v in enumerate(cur_tarry_state) if v == terry_value])
                return '' if step_done else self.step(uistate, i+1)
        print('algorithm end', *uistate.cur_xy)
        values = self.new_values()
        return ''
        

#algorithm = RighthandAlgorithm
algorithm = TarryAlgorithm
try:
    values
except NameError:
    values = algorithm.new_values()
    
def move_L(state, x1, y1, x2, y2, nbridges):
    #print(state, x1, y1, x2, y2, nbridges)
    return state
    
def move_M(state, x, y):
    #print(state, x, y)
    return state
    
def tool(ui, state):
    #print_cdata(ui[0])
    #print_cdata(state[0])
    #print_cdata(state[0].islands, state[0].n_islands)
    
    #_locals = locals()
    #local = globals()
    #local.update(_locals)
    #import code
    #code.interact(banner='', local=local)
    
    uistate = UiState(ui, state)
    try:
        movestr = ensure_valid(uistate)
    except:
        values.direction = 0
        raise
    if movestr is not None:
        return movestr
    try:
        movestr = algorithm.step(uistate)
    except:
        values.direction = 0
        raise
    return movestr
    
    

