#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2019  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import IntEnum
from .common import ffi, print_cdata, printd, toolfunc


ffi.cdef("""
//from tents.c


//#define DIFFLIST(A) \
//    A(EASY,Easy,e) \
//    A(TRICKY,Tricky,t)
//#define ENUM(upper,title,lower) DIFF_ ## upper,
//#define TITLE(upper,title,lower) #title,
//#define ENCODE(upper,title,lower) #lower
//#define CONFIG(upper,title,lower) ":" #title
//enum { DIFFLIST(ENUM) DIFFCOUNT };
//static char const *const tents_diffnames[] = { DIFFLIST(TITLE) };
//static char const tents_diffchars[] = DIFFLIST(ENCODE);
//#define DIFFCONFIG DIFFLIST(CONFIG)

enum {
    COL_BACKGROUND,
    COL_GRID,
    COL_GRASS,
    COL_TREETRUNK,
    COL_TREELEAF,
    COL_TENT,
    COL_ERROR,
    COL_ERRTEXT,
    COL_ERRTRUNK,
    NCOLOURS
};

enum { BLANK, TREE, TENT, NONTENT, MAGIC };

struct game_params {
    int w, h;
    int diff;
};

struct numbers {
    int refcount;
    int *numbers;
};

struct game_state {
    game_params p;
    char *grid;
    struct numbers *numbers;
    int completed, used_solve;
};


struct game_ui {
    int dsx, dsy;                      /* coords of drag start */
    int dex, dey;                      /* coords of drag end */
    int drag_button;                   /* -1 for none, or a button code */
    int drag_ok;                       /* dragged off the window, to cancel */

    int cx, cy, cdisp;                 /* cursor position, and ?display. */
};
""")

class UiState:
    def __init__(self, ui, state):
        self.ui = ui
        self.state = state
        
    class Field(IntEnum): empty, tree, tent, grass = 0, 1, 2, 3
    fieldchars = 'B_TN'
    
    width = property(lambda self: self.state.p.w)
    height = property(lambda self: self.state.p.h)
    
    def coords(self):
        for y in range(self.height):
            for x in range(self.width):
                yield x, y
                
    grid = lambda self,x,y: self.Field(self.state.grid[x+y*self.width][0])
    row = lambda self,y: ((X,y,self.grid(X,y)) for X in range(self.width))
    col = lambda self,x: ((x,Y,self.grid(x,Y)) for Y in range(self.height))
    numbersw = lambda self,x: self.state.numbers[0].numbers[x]
    numbersh = lambda self,y: self.state.numbers[0].numbers[self.width+y]
    
    @property
    def cur_xy(self):
        return self.ui.cx, self.ui.cy
    @cur_xy.setter
    def cur_xy(self, xy):
        self.ui.cx, self.ui.cy = xy
        
    def movestr(self, field, x, y):
        if field == field.tree:
            return None
        f = self.fieldchars[field]
        return f'{f}{x},{y}'
        
    def fields_nontree(self):
        for x, y in self.coords():
            f = self.grid(x, y)
            if f != self.Field.tree:
                yield f, x, y
                
    def fields_empty(self):
        for x, y in self.coords():
            f = self.grid(x, y)
            if f == self.Field.empty:
                yield f, x, y
                
    @staticmethod
    def cycle(f):
        return [f.grass, f.tree, f.empty, f.tent][f]
        
    def around(self, x, y):
        for Y in range(y-1, y+2):
            if 0 <= Y < self.height:
                for X in range(x-1, x+2):
                    if 0 <= X < self.width:
                        if (X,Y) != (x,y):
                            f = self.grid(X,Y)
                            yield X, Y, f, abs(X-x) + abs(Y-y) < 2
                            
    def _is_tent_in_rowcol(self, x, y, roworcol, number):
        f = self.Field
        field_to_matidx = {f.empty:0, f.tent:1, f.grass:2, f.tree:2}
        trans_mat = [   #gr/tr0 tent1  grhit2    tehit3
                        [(1,1), (0,0), (1,3),    (0,2)],     #f.empty
                        [(1,1), (0,1), (1,1),    (0,None)],  #f.tent
                        [(0,0), (0,0), (0,None), (0,0)],     #f.grass, f.tree
                    ]
        def trans_func(old_s, f, hit):
            n, new_s = trans_mat[field_to_matidx[f]][old_s]
            if hit:
                new_s = 3 if new_s == 1 else None
            return n, new_s
        maxn = 0
        s_trans = 0
        for X, Y, f in roworcol:
            n, s_trans = trans_func(s_trans, f, x==X and y==Y)
            if s_trans is None:  return False
            maxn += n
        n, s_trans = trans_func(s_trans, f.grass, False)
        assert n == 0
        if s_trans is None:  return False
        return maxn == number
        
    def _get_field_under_tree(self, x, y, treex, treey):
        result = self.Field.tent
        for X, Y, f, direct in self.around(treex, treey):
            if direct and (X,Y) != (x,y):
                if x==X or y==Y:
                    if f == f.empty:
                        return None
                    elif f == f.tent:
                        F = self._solve_from_around(X, Y, exclude=(treex,treey))
                        if F is None:
                            return None
                        result = F
                else:
                    if f not in [f.grass, f.tree]:
                        return None
        return result
        
    def _solve_from_around(self, x, y, *, exclude=None):
        around = self.Field.grass
        for X, Y, f, direct in self.around(x, y):
            if (X,Y) == exclude:
                continue
            if f == f.tent:
                return f.grass
            if direct and f == f.tree:
                F = self._get_field_under_tree(x, y, X, Y)
                if F is None:
                    around = None
                elif F == F.tent:
                    return F
        return around
        
    _tents_in_rowcol = lambda self,roworcol,*xy:  sum(1 for x,y,f in roworcol if f == f.tent and (x,y)!=xy)
    def _solve_grass_in_finished_row(self, x, y):
        return self.Field.grass if self._tents_in_rowcol(self.row(y),x,y) == self.numbersh(y) else None
    def _solve_grass_in_finished_col(self, x, y):
        return self.Field.grass if self._tents_in_rowcol(self.col(x),x,y) == self.numbersw(x) else None
        
    def _solve_tent_in_row(self, x, y):
        return self.Field.tent if self._is_tent_in_rowcol(x, y, self.row(y), self.numbersh(y)) else None
    def _solve_tent_in_col(self, x, y):
        return self.Field.tent if self._is_tent_in_rowcol(x, y, self.col(x), self.numbersw(x)) else None
        
    def try_solve(self, f_cur, x, y):
        if f_cur == f_cur.tree:
            self.move_cur(x, y)
            return ''
        for name in dir(self):
            if name.startswith('_solve_'):
                f = getattr(self, name)(x, y)
                if f is not None:
                    if f != f_cur:
                        return self.movestr(f, x, y)
        self.move_cur(x, y)
        return ''
        
    def move_cur(self, x, y):
        def test_skip_row(rangex):
            Y = min(y+1, self.height-1)
            for X in rangex:
                if self.grid(X,y) == self.Field.empty or self.grid(X,Y) == self.Field.empty:
                    return False
            return True
        if y%2 == 0:
            if test_skip_row(range(x+1, self.width)):
                y += 1
            else:
                x += 1
        else:
            if test_skip_row(range(0, x)):
                y += 1
            else:
                x -= 1
        if y >= self.height:
            x = y = 0
        self.cur_xy = x, y
        
def move(mtype, state, x, y):
    return state
    
@toolfunc('Rotate Fields')
def tool(ui, state):
    uistate = UiState(ui, state)
    return ';'.join([uistate.movestr(uistate.cycle(f),x,y) for f,x,y in uistate.fields_nontree()])
    
@toolfunc('Fill with grass')
def tool(ui, state):
    uistate = UiState(ui, state)
    return ';'.join([uistate.movestr(f.grass,x,y) for f,x,y in uistate.fields_empty()])
    
@toolfunc('Fill with tents')
def tool(ui, state):
    uistate = UiState(ui, state)
    return ';'.join([uistate.movestr(f.tent,x,y) for f,x,y in uistate.fields_empty()])
    
@toolfunc('Solve step')
def tool(ui, state):
    if not ui.cdisp:
        ui.cdisp = 1
        return ''
    uistate = UiState(ui, state)
    x, y = uistate.cur_xy
    f = uistate.grid(x, y)
    return uistate.try_solve(f,x,y)
    
@toolfunc('debug')
def tool(ui, state):
    #print_cdata(ui[0])
    #print_cdata(state[0])
    
    uistate = UiState(ui, state)
    
    _locals = locals()
    local = globals()
    local.update(_locals)
    import code
    code.interact(banner='', local=local)
    
    return None
    

