#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright ©  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import enum
from contextlib import suppress
from .common import ffi, print_cdata, printd, toolfunc


# from tracks.c:
ffi.cdef("""
//#define DIFFLIST(A) \
//    A(EASY,Easy,e) \
//    A(TRICKY,Tricky,t)

//#define ENUM(upper,title,lower) DIFF_ ## upper,
//#define TITLE(upper,title,lower) #title,
//#define ENCODE(upper,title,lower) #lower
//#define CONFIG(upper,title,lower) ":" #title
//enum { DIFFLIST(ENUM) DIFFCOUNT };
static char const *const tracks_diffnames[] = { DIFFLIST(TITLE) };
static char const tracks_diffchars[] = DIFFLIST(ENCODE);
//#define DIFFCONFIG DIFFLIST(CONFIG)

struct game_params {
    int w, h, diff, single_ones;
};

static const struct game_params tracks_presets[] = {
    {8, 8, DIFF_EASY, 1},
    {8, 8, DIFF_TRICKY, 1},
    {10, 8, DIFF_EASY, 1},
    {10, 8, DIFF_TRICKY, 1 },
    {10, 10, DIFF_EASY, 1},
    {10, 10, DIFF_TRICKY, 1},
    {15, 10, DIFF_EASY, 1},
    {15, 10, DIFF_TRICKY, 1},
    {15, 15, DIFF_EASY, 1},
    {15, 15, DIFF_TRICKY, 1},
};

#define R 1
#define U 2
#define L 4
#define D 8

//#define MOVECHAR(m) ((m==R)?'R':(m==U)?'U':(m==L)?'L':(m==D)?'D':'?')

//#define DX(d) ( ((d)==R) - ((d)==L) )
//#define DY(d) ( ((d)==D) - ((d)==U) )

//#define F(d) (((d << 2) | (d >> 2)) & 0xF)
//#define C(d) (((d << 3) | (d >> 1)) & 0xF)
//#define A(d) (((d << 1) | (d >> 3)) & 0xF)

//#define LR (L | R)
//#define RL (R | L)
//#define UD (U | D)
//#define DU (D | U)
//#define LU (L | U)
//#define UL (U | L)
//#define LD (L | D)
//#define DL (D | L)
//#define RU (R | U)
//#define UR (U | R)
//#define RD (R | D)
//#define DR (D | R)
#define ALLDIR 15
#define BLANK 0
#define UNKNOWN 15

int nbits[] = { 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4 };

/* square grid flags */
#define S_TRACK 1     /* a track passes through this square (--> 2 edges) */
#define S_NOTRACK 2   /* no track passes through this square */
#define S_ERROR 4
#define S_CLUE 8
#define S_MARK 16

#define S_TRACK_SHIFT   16 /* U/D/L/R flags for edge track indicators */
#define S_NOTRACK_SHIFT 20 /* U/D/L/R flags for edge no-track indicators */

/* edge grid flags */
#define E_TRACK 1     /* a track passes through this edge */
#define E_NOTRACK 2   /* no track passes through this edge */

struct numbers {
    int refcount;
    int *numbers;     /* sz w+h */
    int row_s, col_s; /* stations: TODO think about multiple lines
                         (for bigger grids)? */
};

//#define INGRID(state, gx, gy) ((gx) >= 0 && (gx) < (state)->p.w && \
//                               (gy) >= 0 && (gy) < (state)->p.h)

struct game_state {
    game_params p;
    unsigned int *sflags;       /* size w*h */
    struct numbers *numbers;
    int *num_errors;            /* size w+h */
    int completed, used_solve, impossible;
};

#define NDIRS 4
const unsigned int dirs_const[] = { U, D, L, R };

struct tracks_neighbour_ctx {
    game_state *state;
    int i, n, neighbours[4];
};

struct game_ui {
    int dragging, clearing, notrack;
    int drag_sx, drag_sy, drag_ex, drag_ey; /* drag start and end grid coords */
    int clickx, clicky;    /* pixel position of initial click */

    int curx, cury;        /* grid position of keyboard cursor; uses half-size grid */
    int cursor_active;     /* TRUE iff cursor is shown */
};

#define DS_DSHIFT 4     /* R/U/L/D shift, for drag-in-progress flags */

//#define DS_ERROR (1 << 8)
//#define DS_CLUE (1 << 9)
//#define DS_NOTRACK (1 << 10)
//#define DS_FLASH (1 << 11)
//#define DS_CURSOR (1 << 12) /* cursor in square (centre, or on edge) */
//#define DS_TRACK (1 << 13)
//#define DS_CLEARING (1 << 14)

#define DS_NSHIFT 16    /* R/U/L/D shift, for no-track edge flags */
#define DS_CSHIFT 20    /* R/U/L/D shift, for cursor-on-edge */

struct game_drawstate {
    int sz6;
    int started;

    int w, h, sz;
    unsigned int *flags, *flags_drag;
    int *num_errors;
};

""")

S_TRACK = 1
S_NOTRACK = 2
S_ERROR = 4
S_CLUE = 8
S_MARK = 16

S_TRACK_SHIFT   = 16
S_NOTRACK_SHIFT = 20


track_mask = (0xf << S_TRACK_SHIFT) | S_TRACK
no_mask = (0xf << S_NOTRACK_SHIFT) | S_NOTRACK
full_mask = track_mask | S_NOTRACK

class FillMode(enum.Flag): none, track, notrack, both, uniq, x, y, xy, restrictive = 0, 1, 2, 3, 4, 8, 16, 24, 32
class NoIsle(Exception):  pass

class State:
    def __init__(self, state):
        self.w = state[0].p.w
        self.h = state[0].p.h
        self.numbersx = state[0].numbers[0].numbers[0:self.w]
        self.numbersy = state[0].numbers[0].numbers[self.w:self.w+self.h]
        self._sflags = state[0].sflags
        
    sflags = lambda self, x,y: self._sflags[y*self.w+x]
    
    def movestr(self, s, x, y):
        assert 0 <= x < self.w, (x, self.w)
        assert 0 <= y < self.h, (y, self.h)
        return f'{s}{x},{y}'
    def movestr_block(self, x, y):  return self.movestr('TS', x, y)
    def movestr_no(self, x, y):     return self.movestr('NS', x, y)
    
    def _naives_füllen_xy(self, w, h, numbers, fillmode, xyfunc, moves_block, moves_no):
        _track_mask = 0xf << S_TRACK_SHIFT if fillmode & fillmode.restrictive else track_mask
        for a in range(w):
            sum_track = 0
            sum_no = 0
            sum_empty = 0
            ny = numbers[a]
            for b in range(h):
                x,y = xyfunc(a,b)
                f = self.sflags(x,y)
                if not f & full_mask:
                    sum_empty += 1
                elif f & _track_mask:
                    sum_track += 1
                elif f & S_NOTRACK:
                    sum_no += 1
                else:
                    sum_empty += 1
            assert sum_empty + sum_track + sum_no == h, (sum_empty, sum_track, sum_no, h)
            guard_track = fillmode & fillmode.track or fillmode & fillmode.uniq and sum_track + sum_empty == ny
            guard_no = fillmode & fillmode.notrack or fillmode & fillmode.uniq and sum_no + sum_empty == h - ny
            for b in range(h):
                x,y = xyfunc(a,b)
                f = self.sflags(x,y)
                if not f & full_mask:
                    assert sum_empty > 0
                    if sum_track < ny and guard_track:
                        if (x, y) in moves_no:
                            moves_no.remove((x, y))
                        else:
                            if (x, y) not in moves_block:
                                moves_block.append((x, y))
                            sum_track += 1
                    elif sum_no < h-ny and guard_no:
                        if (x, y) in moves_block:
                            moves_block.remove((x, y))
                        else:
                            if (x, y) not in moves_no:
                                moves_no.append((x, y))
                            sum_no += 1
                    sum_empty -= 1
        
    def naives_füllen(self, fillmode):
        moves_block = []
        moves_no = []
        if fillmode & fillmode.x:
            self._naives_füllen_xy(self.h, self.w, self.numbersy, fillmode, lambda x,y: (y,x),
                                            moves_block, moves_no)
        if fillmode & fillmode.y:
            self._naives_füllen_xy(self.w, self.h, self.numbersx, fillmode, lambda x,y: (x,y),
                                            moves_block, moves_no)
        moves = [self.movestr_block(x,y) for x,y in moves_block]
        moves += [self.movestr_no(x,y) for x,y in moves_no]
        return ';'.join(moves) or None
        
    def extend_isle(self, x1, y1, x2, y2):
        def cnt_sflags(_x1, _y1, _x2, _y2, mask):
            nonlocal x1, y1, x2, y2
            if not (0 <= _x1 <= _x2 <= self.w and 0 <= _y1 <= _y2 <= self.h):
                return False
            sumn = sume = 0
            for y in range(_y1, _y2):
                for x in range(_x1, _x2):
                    if self.sflags(x,y) & full_mask:
                        if self.test_sflags(x, y, mask, 1):
                            raise NoIsle
                        else:
                            sumn += 1
                    else:
                        sume += 1
            assert sumn + sume == (_x2 - _x1) * (_y2 - _y1)
            if sumn == 0:
                x1 = min(x1, _x1)
                y1 = min(y1, _y1)
                x2 = max(x2, _x2)
                y2 = max(y2, _y2)
                return True
            if sume == 0:
                return False
            raise NoIsle
        changed = True
        while changed:
            changed  = cnt_sflags(x2,   y1,   x2+1, y2,   0b0100)  # rechts
            changed |= cnt_sflags(x1,   y1-1, x2,   y1,   0b1000)  # oben
            changed |= cnt_sflags(x1-1, y1,   x1,   y2,   0b0001)  # links
            changed |= cnt_sflags(x1,   y2,   x2,   y2+1, 0b0010)  # unten
        return [(x,y) for y in range(y1, y2) for x in range(x1, x2)]
        
    def test_sflags(self, x, y, f, res):
        if x < 0:  return 0
        if x >= self.w:  return 0
        if y < 0:  return 0
        if y >= self.h:  return 0
        sf = self.sflags(x,y)
        if (sf >> S_NOTRACK_SHIFT) & f:  return 0
        if sf & S_NOTRACK:  return 0
        if (sf >> S_TRACK_SHIFT) & f:  return res
        if bin((sf >> S_TRACK_SHIFT) & 0b1111).count('1') >= 2:  return 0
        if sf & S_TRACK:  return res
        return res
        
@toolfunc('Eindeutiges Füllen')
def tool(ui, state):
    return State(state).naives_füllen(FillMode.uniq|FillMode.xy)
    
@toolfunc('Inseln füllen')
def tool(ui, state):
    state = State(state)
    moves = []
    for y in range(state.h):
        for x in range(state.w):
            if (x,y) in moves:
                continue
            if state.sflags(x,y) & full_mask:
                continue
            with suppress(NoIsle):
                moves += state.extend_isle(x, y, x+1, y+1)
    return ';'.join(state.movestr_no(x,y) for x,y in moves) or None
    
@toolfunc('Kleine (Halb-)Inseln füllen')
def tool(ui, state):
    state = State(state)
    moves = []
    changed = True
    while changed:
        changed = False
        for y in range(state.h):
            for x in range(state.w):
                if (x,y) in moves:
                    continue
                if state.sflags(x,y) & full_mask:
                    continue
                cnt = 0
                for a, b, l, m in (x+1,y,x+1-state.w,0b0100), (x,y-1,y,0b1000), (x-1,y,x,0b0001), (x,y+1,y+1-state.h,0b0010):
                    if l == 0 or not state.test_sflags(a, b, m, 1) or (a,b) in moves:
                        cnt += 1
                if cnt > 2:
                    moves.append((x,y))
                    changed = True
    return ';'.join(state.movestr_no(x,y) for x,y in moves) or None
    
@toolfunc('Eindeutige Schienen verlegen')
def tool(ui, state):
    state = State(state)
    moves = set()
    for y in range(state.h):
        for x in range(state.w):
            f = state.sflags(x,y)
            if not f & track_mask:
                continue
            strack = (f >> S_TRACK_SHIFT) & 0b1111
            if bin(strack).count('1') > 1:
                continue
            ttrack = state.test_sflags(x+1, y, 0b0100, 0b0001) | \
                     state.test_sflags(x, y-1, 0b1000, 0b0010) | \
                     state.test_sflags(x-1, y, 0b0001, 0b0100) | \
                     state.test_sflags(x, y+1, 0b0010, 0b1000)
            if (strack | ttrack) != ttrack:
                continue
            if ttrack == 0b0011:
                moves.add(state.movestr('TR',x,y))
                moves.add(state.movestr('TD',x,y-1))
            elif ttrack == 0b0101:
                moves.add(state.movestr('TR',x,y))
                moves.add(state.movestr('TR',x-1,y))
            elif ttrack == 0b0110:
                moves.add(state.movestr('TD',x,y-1))
                moves.add(state.movestr('TR',x-1,y))
            elif ttrack == 0b1001:
                moves.add(state.movestr('TR',x,y))
                moves.add(state.movestr('TD',x,y))
            elif ttrack == 0b1010:
                moves.add(state.movestr('TD',x,y-1))
                moves.add(state.movestr('TD',x,y))
            elif ttrack == 0b1100:
                moves.add(state.movestr('TR',x-1,y))
                moves.add(state.movestr('TD',x,y))
            else:
                continue
    return ';'.join(moves) or None
    
@toolfunc('Restriktives Füllen')
def tool(ui, state):
    return State(state).naives_füllen(FillMode.uniq|FillMode.xy|FillMode.restrictive)
    
@toolfunc('Naives Füllen x')
def tool(ui, state):
    return State(state).naives_füllen(FillMode.both|FillMode.x)
    
@toolfunc('Naives Füllen y')
def tool(ui, state):
    return State(state).naives_füllen(FillMode.both|FillMode.y)
    
@toolfunc('Naives Füllen (Track) x')
def tool(ui, state):
    return State(state).naives_füllen(FillMode.track|FillMode.x)
    
@toolfunc('Naives Füllen (Track) y')
def tool(ui, state):
    return State(state).naives_füllen(FillMode.track|FillMode.y)
    
@toolfunc('inspect args')
def tool(ui, state):
    print_cdata(ui[0])
    print_cdata(state[0])
    print_cdata(state[0].numbers[0])
    w = state[0].p.w
    h = state[0].p.h
    print(*[state[0].numbers[0].numbers[n] for n in range(w+h)])
    #_locals = locals()
    #local = globals()
    #local.update(_locals)
    #import code
    #code.interact(banner='', local=local)
    
def premove(pystate, mchar1, mchar2x, y):
    mchar2 = mchar2x[0]
    x = mchar2x[1:]
    if mchar1 == 'N' and mchar2 != 'S':
        return False
    return True
    
def move(*args):
    pass
    

