#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2019  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
import math
from .common import ffi, print_cdata, printd, toolfunc


with Path(__file__).with_name('tree234.h_').open('rt', encoding='utf8') as _fn:
    ffi.cdef(_fn.read())
with Path(__file__).with_name('tree234.c_').open('rt', encoding='utf8') as _fn:
    ffi.cdef(_fn.read())
    del _fn
    
ffi.cdef("""
//from tents.c

#define CIRCLE_RADIUS 6
//#define DRAG_THRESHOLD (CIRCLE_RADIUS * 2)
#define PREFERRED_TILESIZE 64

//#define FLASH_TIME 0.30F
//#define ANIM_TIME 0.13F
//#define SOLVEANIM_TIME 0.50F

enum {
    COL_SYSBACKGROUND,
    COL_BACKGROUND,
    COL_LINE,
//#ifdef SHOW_CROSSINGS
//    COL_CROSSEDLINE,
//#endif
    COL_OUTLINE,
    COL_POINT,
    COL_DRAGPOINT,
    COL_NEIGHBOUR,
    COL_FLASH1,
    COL_FLASH2,
    NCOLOURS
};

typedef struct point {
    /*
     * Points are stored using rational coordinates, with the same
     * denominator for both coordinates.
     */
    long x, y, d;
} point;

typedef struct edge {
    /*
     * This structure is implicitly associated with a particular
     * point set, so all it has to do is to store two point
     * indices. It is required to store them in the order (lower,
     * higher), i.e. a < b always.
     */
    int a, b;
} edge;

struct game_params {
    int n;			       /* number of points */
};

struct graph {
    int refcount;		       /* for deallocation */
    tree234 *edges;		       /* stores `edge' structures */
};

struct game_state {
    game_params params;
    int w, h;			       /* extent of coordinate system only */
    point *pts;
//#ifdef SHOW_CROSSINGS
//    int *crosses;		       /* mark edges which are crossed */
//#endif
    struct graph *graph;
    int completed, cheated, just_solved;
};

//typedef struct {
//    long hi;
//    unsigned long lo;
//} int64;

/*
 * Our solutions are arranged on a square grid big enough that n
 * points occupy about 1/POINTDENSITY of the grid.
 */
#define POINTDENSITY 3
#define MAXDEGREE 4
//#define COORDLIMIT(n) squarert((n) * POINTDENSITY)

//typedef struct vertex {
//    int param;
//    int vindex;
//} vertex;

struct game_ui {
    int dragpoint;		       /* point being dragged; -1 if none */
    point newpoint;		       /* where it's been dragged to so far */
    int just_dragged;		       /* reset in game_changed_state */
    int just_moved;		       /* _set_ in game_changed_state */
    float anim_length;
};

""")


@toolfunc('Strecken horiz/vert')
def tool(ui, state):
    w = state[0].w
    h = state[0].h
    xmin = xmax = w/2
    ymin = ymax = h/2
    
    for i in range(state[0].params.n):
        p = state[0].pts[i]
        if p.x/p.d < xmin:  xmin = p.x/p.d
        if p.x/p.d > xmax:  xmax = p.x/p.d
        if p.y/p.d < ymin:  ymin = p.y/p.d
        if p.y/p.d > ymax:  ymax = p.y/p.d
        print(f'  {i}: {p.x},{p.y},{p.d} ({p.x/p.d},{p.y/p.d})')
    def get_moves():
        for i in range(state[0].params.n):
            p = state[0].pts[i]
            x = int((p.x-xmin*p.d)*(w-1)//(xmax-xmin) + p.d/2)
            y = int((p.y-ymin*p.d)*(h-1)//(ymax-ymin) + p.d/2)
            if i == ui[0].dragpoint:
                ui[0].newpoint.x = x
                ui[0].newpoint.y = y
                ui[0].newpoint.d = p.d
            yield f'P{i}:{x},{y}/{p.d}'
    return ';'.join(get_moves())
    
@toolfunc('Strecken zentr')
def tool(ui, state):
    w = state[0].w
    h = state[0].h
    cx = w/2
    cy = h/2
    
    pinfo = []
    d2max = 0
    d2new = (min(w,h))/2
    #d2new *= d2new
    for i in range(state[0].params.n):
        p = state[0].pts[i]
        dx = p.x/p.d - cx
        dy = p.y/p.d - cy
        d2 = math.sqrt(dx*dx + dy*dy)
        pinfo.append((i,p,dx,dy,d2))
        if d2max < d2:  d2max = d2
    res = ''
    #d2max = math.sqrt(d2max)
    print(f'd2max: {d2max}  d2new: {d2new}')
    for i,p,dx,dy,d2 in pinfo:
        d2t = math.atan(d2/d2new) * 4 / math.pi * d2new
        x = int((cx + dx * d2t / d2) * p.d)
        y = int((cy + dy * d2t / d2) * p.d)
        print(f'P{i}:{x},{y}/{p.d};  {p.x},{p.y}  {w*p.d},{h*p.d}')
        res += f'P{i}:{x},{y}/{p.d};'
        if i == ui[0].dragpoint:
            ui[0].newpoint.x = x
            ui[0].newpoint.y = y
            ui[0].newpoint.d = p.d
    return res
    
def get_edges1(state):
    def _gen_edges(node):
        for k in node.kids:
            if not k:  break
            yield from _gen_edges(k[0])
        for e in node.elems:
            if not e:  break
            e = ffi.cast('edge*', e)
            yield e.a, e.b
    yield from _gen_edges(state[0].graph[0].edges[0].root[0])
    
def get_edges2(state):
    for a, b in get_edges1(state):
        yield a, b
        yield b, a
    
def _center_point_helper(ui, state, func):
    i = ui[0].dragpoint
    if i < 0:
        return
    p = state[0].pts[i]
    p, sx, sy, n = func(p)
    for a, b in get_edges2(state):
        if a == i:
            q = state[0].pts[b]
            sx += q.x/q.d
            sy += q.y/q.d
            n += 1
    sx = int(sx / n * p.d)
    sy = int(sy / n * p.d)
    newpoint = ui[0].newpoint
    newpoint.x = sx
    newpoint.y = sy
    newpoint.d = p.d
    return f'P{i}:{sx},{sy}/{p.d}'
    
@toolfunc('Zentriere Punkt weich')
def tool(ui, state):
    return _center_point_helper(ui, state, lambda p: (p, p.x/p.d, p.y/p.d, 1))
    
@toolfunc('Zentriere Punkt hart')
def tool(ui, state):
    return _center_point_helper(ui, state, lambda p: (p, 0, 0, 0))
    
@toolfunc('inspect args')
def tool(ui, state):
    print_cdata(ui[0])
    print_cdata(state)
    #print_cdata(state[0].pts, state[0].params.n)
    
    #uistate = UiState(ui, state)
    
    #_locals = locals()
    #local = globals()
    #local.update(_locals)
    #import code
    #code.interact(banner='', local=local)
    
def move(*args):
    pass
    

