# -*- coding: utf-8 -*-

#  Copyright © 2016  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import importlib
from cffi import FFI
from pathlib import Path

ffi = FFI()
with Path(__file__).with_name('puzzles.h_').open('rt', encoding='utf8') as _fn:
    ffi.cdef(_fn.read())
    del _fn
    

def import_module(modulename):
    modulename = modulename.replace(' ', '_')
    print('  import', modulename)
    return importlib.import_module(modulename)
    
def reload(module):
    #print('  reload', module.__name__)
    return importlib.reload(module)
    
    
def print_cdata(data, *datarange, indent='', name=None):
    if not isinstance(data, FFI.CData):
        if name is None:
            name = type(data).__name__
        kind = 'python'
    else:
        datatype = ffi.typeof(data)
        if name is None:
            name = datatype.cname
        kind = datatype.kind
    print('{}{!r} {} ({})'.format(indent, name, data, kind))
    if kind == 'struct':
        for name, unused in datatype.fields:
            value = getattr(data, name)
            print_cdata(value, indent=indent+'  ', name=name)
    elif kind == 'array':
        if not datarange:
            datarange = (len(data),)
        for i in range(*datarange):
            try:
                value = data[i]
            except IndexError:
                break
            print_cdata(value, indent=indent+'  ', name='[{}]'.format(i))
    elif kind == 'pointer' and datarange:
        for i in range(*datarange):
            value = data[i]
            print_cdata(value, indent=indent+'  ', name='[{}]'.format(i))
            
def printd(text=''):
    import inspect
    f = inspect.currentframe().f_back
    print(f'dbg:{f.f_lineno}:{f.f_code.co_name}:', text)
    
    
try:
    pystates
except NameError:
    pystates = []
    pystatepos = 0
assert not pystates and pystatepos==0 or 0 <= pystatepos < len(pystates), (pystatepos, len(pystates))
    
tools = []
def toolfunc(title):
    def deco_func(func):
        tools.append((title, func))
        return func
    return deco_func
    
def menu(module, ui, toolnumber):
    try:
        return tools[toolnumber][0]
    except AttributeError:
        return 'Tool' if toolnumber == 0 else None
    except IndexError:
        return None
    
def game(module, ui, state, gtype):
    #print('game:', gtype)
    ui = ffi.cast("game_ui*", ui)
    state = ffi.cast("game_state*", state)
    
    global pystatepos
    if gtype == 'new':
        pystates.clear()
        pystatepos = 0
    elif gtype == 'restart' and pystates:
        pystates.append(pystates[0])
        pystatepos += 1
    elif gtype == 'undo' and pystatepos:
        pystatepos -= 1
    elif gtype == 'redo' and pystatepos < len(pystates)-1:
        pystatepos += 1
    assert not pystates and pystatepos==0 or 0 <= pystatepos < len(pystates), (pystatepos, len(pystates))
        
def tool(module, toolnumber, ui, state):
    ui = ffi.cast("game_ui*", ui)
    state = ffi.cast("game_state*", state)
    pystatec = pystates[pystatepos] if pystates else None
    if pystatec is not None:
        module.values = pystatec
    try:
        tool = tools[toolnumber][1]
    except IndexError:
        if toolnumber == 0:
            tool = module.tool
        else:
            raise
    return tool(ui, state)
    
def premove(module, ui, state, movestr):
    pystatec = pystates[pystatepos] if pystates else None
    if pystatec is not None:
        import copy
        pystatec = copy.deepcopy(pystatec)
    def convert_arg(a):
        try:
            return int(a)
        except ValueError:
            return a
    res = True
    for mstr in movestr.split(';'):
        if not mstr:
            continue
        mtype = mstr[0]
        margs = [convert_arg(a) for a in mstr[1:].split(',')]
        try:
            func = getattr(module, 'premove')
            margs = [mtype] + margs
        except AttributeError:
            func = lambda pystatec, *margs: print(f'not implemented: premove{(mtype,)+margs}')
        if not func(pystatec, *margs):
            res = False
    return res
    
def move(module, ui, state, movestr):
    #ui = ffi.cast("game_ui*", ui)
    #state = ffi.cast("game_state*", state)
    global pystatepos
    pystatec = pystates[pystatepos] if pystates else None
    if pystatec is not None:
        import copy
        pystatec = copy.deepcopy(pystatec)
    def convert_arg(a):
        try:
            return int(a)
        except ValueError:
            return a
    for mstr in movestr.split(';'):
        if not mstr:
            continue
        mtype = mstr[0]
        mfunc = 'move_' + mtype
        margs = [convert_arg(a) for a in mstr[1:].split(',')]
        try:
            func = getattr(module, mfunc)
        except AttributeError:
            try:
                func = getattr(module, 'move')
                margs = [mtype] + margs
            except AttributeError:
                func = lambda pystatec, *margs: print(f'not implemented: {mfunc}{margs} or move{(mtype,)+margs}')
        pystatec = func(pystatec, *margs)
    if pystates:
        pystatepos += 1
        del pystates[pystatepos:]
    pystates.append(pystatec)
    assert not pystates and pystatepos==0 or 0 <= pystatepos < len(pystates), (pystatepos, len(pystates))
    
def call(module, func, args):
    #print(__name__, 'call', module, func, args)
    return globals()[func](module, *args)
    
    

